#
# Author: Sebastien Verel
#         Univ. du Littoral Côte d'Opale
# Date: 
#   2018/11/20
#
# Licence: free
#

# create data for basic linear model regression
create_data <- function(n, outname) {
  x = runif(n = n, min = -10, max = 10)
  noice <- rnorm(n, mean = 0, sd = 1)
  y = -2 + 0.5 *x + noice
  
  # -1.925940 + 0.491260 * x
  write.table(data.frame(y = y, x = x), file = outname, sep = " ", row.names = FALSE)
}

# Mean Square Error (MSE) function of the linear model computed on data.frame df
mse.fun <- function(df) {
  f <- function(theta) {
    return(0.5 * mean( (theta[1] + theta[2] * df$x - df$y)^2 ) )
  }
  
  return(f)
}

mse.gradient.fun <- function(df) {
  g <- function(theta) {
    g1 <-  mean(theta[1] + theta[2] * df$x - df$y)
    g2 <-  mean(df$x * (theta[1] + theta[2] * df$x - df$y))
    return(c(g1, g2))
  }
  
  return(g)
}

gradient.descent <- function(gradientf, x0, sigma, niter) {
  x <- x0

  iter <- 0  
  while (iter < niter) {
    w <- - gradientf(x)
    x <- x + sigma * w
    
    iter <- iter + 1
  }
  
  return(list(x=x, dyn=c()))
}

mse.gradient_i.fun <- function(df) {
  g <- function(i, theta) {
    x <- df[i,]$x
    y <- df[i,]$y
    g1 <-  theta[1] + theta[2] * x - y
    g2 <-  x * (theta[1] + theta[2] * x - y)
    return(c(g1, g2))
  }
  
  return(g)
}

stochastic.gradient <- function(data_size, gradientf_i, x0, sigma, niter) {
  x <- x0

  index <- sample(1:data_size)
  
  iter <- 0  
  while (iter < niter) {
    
    for(i in index) {
      w <- - gradientf_i(i, x)
      x <- x + sigma * w
    }

    iter <- iter + 1
  }
  
  return(list(x=x, dyn=c()))
}


main <- function() {
  #create_data(n = 500, "data1.csv")
  
  # read data
  df1 <- read.table("data1.csv", header = TRUE, sep = " ")
  
  # plot data
  plot(y ~ x, data = df1)
  
  # mse function
  mse <- mse.fun(df1)
  
  # basic initial parameters: intercept theta[1] = mean(y), and slope theta[2] = 0
  theta <- c(mean(df1$y), 0)
  
  # plot initial model, print initial quality
  x <- seq(-10, 10, by = 0.01)
  lines(x, theta[1] + theta[2] * x)
  cat("mse(theta)=", mse(theta))

  # gradient of the mse function
  mse.gradient <- mse.gradient.fun(df1)
  
  # gradient descent with fix step size
  res <- gradient.descent(f = mse, gradientf = mse.gradient, x0 = theta, sigma = 0.01, precision = 1e-3, niter = 1000)
  
  # plot final model
  lines(x, res$x[1] + res$x[2] * x, col = "red", lw = 4)
  cat("mse(theta)=", mse(res$x))
  
  # dynamique
  plot(log(res$dyn), type = "l")
  
  mse.gradient_i <- mse.gradient_i.fun(df1)
  res <-  stochastic.gradient(nrow(df1), mse.gradient_i, x0 = theta, sigma = 0.01, niter = 10)
}

